//
//  ES1Renderer.m
//  Shatter
//
//  Created by Yoshikazu Kuramochi on 11/09/30.
//  Copyright __MyCompanyName__ 2011. All rights reserved.
//

#import "ES1Renderer.h"

@implementation ES1Renderer

const int NUM_FRAGMENTS = 10 * 15;			// 横10, 縦15
const int NUM_VERTICES = 10 * 15 * 2 * 3;	// 横10, 縦15, 1マスに2個の三角形, 三角形の頂点数3

- (void) updateVertices
{
	CFAbsoluteTime time = CFAbsoluteTimeGetCurrent();

	for (int j = 0; j < 15; ++j) {
		for (int i = 0; i < 10; ++i) {
			int k = (j * 10 + i) * (3 + 3 + 3 + 1 + 1);
			
			GLfloat cx, cy, cz;
			if (shattered && time - shatteredTime >= shatterData[k+10]) {
				// 中心の速度
				shatterData[k+3] += accelVec[0] / 500;
				shatterData[k+4] += accelVec[1] / 500;
				shatterData[k+5] += accelVec[2] / 500;

				// 中心の位置
				cx = (shatterData[k   ] += shatterData[k+3] /* * dt*/);
				cy = (shatterData[k+ 1] += shatterData[k+4] /* * dt*/);
				cz = (shatterData[k+ 2] += shatterData[k+5] /* * dt*/);
			} else {
				cx = shatterData[k   ];
				cy = shatterData[k+ 1];
				cz = shatterData[k+ 2];
			}
			
			// 頂点の位置
			k = (j * 10 + i) * 2 * 3 * 3;
			vertices[k   ] = cx - 1/10.0;
			vertices[k+ 1] = cy - 1/15.0;
			vertices[k+ 2] = cz;
			vertices[k+ 3] = cx + 1/10.0;
			vertices[k+ 4] = cy - 1/15.0;
			vertices[k+ 5] = cz;
			vertices[k+ 6] = cx - 1/10.0;
			vertices[k+ 7] = cy + 1/15.0;
			vertices[k+ 8] = cz;
			vertices[k+ 9] = cx + 1/10.0;
			vertices[k+10] = cy + 1/15.0;
			vertices[k+11] = cz;
			vertices[k+12] = cx - 1/10.0;
			vertices[k+13] = cy + 1/15.0;
			vertices[k+14] = cz;
			vertices[k+15] = cx + 1/10.0;
			vertices[k+16] = cy - 1/15.0;
			vertices[k+17] = cz;
		}
	}
}

- (void) resetShatterData
{
	for (int j = 0; j < 15; ++j) {
		for (int i = 0; i < 10; ++i) {
			int k = (j * 10 + i) * (3 + 3 + 3 + 1 + 1);
			
			// 中心の位置
			shatterData[k  ] = -1 + (i+0.5)*2/10.0;
			shatterData[k+1] = -1 + (j+0.5)*2/15.0;
			shatterData[k+2] = 0;
			
			// 中心の速度
			shatterData[k+3] = 0;
			shatterData[k+4] = 0;
			shatterData[k+5] = 0;
			
			// 回転軸
			shatterData[k+6] = 0;
			shatterData[k+7] = 0;
			shatterData[k+8] = 1;
			
			// 回転速度
			shatterData[k+9] = 0;
			
			// 遅延
			shatterData[k+10] = INFINITY;
			
			
			// テクスチャ座標
			k = (j * 10 + i) * 2 * 3 * 2;
			texCoords[k   ] =      i/10.0;
			texCoords[k+ 1] =  1 - j/15.0;
			texCoords[k+ 2] =      (i+1)/10.0;
			texCoords[k+ 3] =  1 - j/15.0;
			texCoords[k+ 4] =      i/10.0;
			texCoords[k+ 5] =  1 - (j+1)/15.0;
			texCoords[k+ 6] =      (i+1)/10.0;
			texCoords[k+ 7] =  1 - (j+1)/15.0;
			texCoords[k+ 8] =      i/10.0;
			texCoords[k+ 9] =  1 - (j+1)/15.0;
			texCoords[k+10] =      (i+1)/10.0;
			texCoords[k+11] =  1 - j/15.0;
		}
	}

	[self updateVertices];
}

-(void) loadTexture
{
//	NSString* file = @"img1.png";
	NSString* file = @"iori.png";
	NSString* path = [[NSBundle mainBundle] pathForResource:file ofType:nil];
	UIImage* image = [UIImage imageWithContentsOfFile:path];
	CGImageRef cgImage = [image CGImage];
	size_t width = CGImageGetWidth(cgImage);
    size_t height = CGImageGetHeight(cgImage);
	CFDataRef data = CGDataProviderCopyData(CGImageGetDataProvider(cgImage));
	
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

#if TARGET_IPHONE_SIMULATOR
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0,
				 GL_RGB, GL_UNSIGNED_BYTE, CFDataGetBytePtr(data));
#else
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
				 GL_BGRA, GL_UNSIGNED_BYTE, CFDataGetBytePtr(data));
#endif

	CFRelease(data);
}

// Create an ES 1.1 context
- (id) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		
		[self loadTexture];
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glGenRenderbuffersOES(1, &depthRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);

		glClearColor(0, 0, 0, 0);
		glClearDepthf(1.0);
		glColor4f(1, 1, 1, 1);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		double tan5 = tan(5.0 / 180.0 * 3.14159265358979323846264);
		double plus = 0.1*tan5;
		double minus = -plus; 
		glFrustumf(minus, plus, minus, plus, 0.1, 1000.0);
		
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0.0, 0.0, -1/tan5);
		
		glBindTexture(GL_TEXTURE_2D, texture);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);


										// 位置(3) + 速度(3) + 回転軸(3) + 回転速度(1) + 遅延(1)
		shatterData = malloc(NUM_FRAGMENTS * (3 + 3 + 3 + 1 + 1) * sizeof(GLfloat));
		vertices = malloc(NUM_VERTICES * 3 * sizeof(GLfloat));
		texCoords = malloc(NUM_VERTICES * 2 * sizeof(GLfloat));

		[self resetShatterData];

		UIAccelerometer* accel = [UIAccelerometer sharedAccelerometer];
		accel.updateInterval = 1/60.0;
		accel.delegate = self;
	}
	
	return self;
}

- (void) render
{
	if (shattered) {
		[self updateVertices];
	}
	
	// This application only creates a single context which is already set current at this point.
	// This call is redundant, but needed if dealing with multiple contexts.
    //[EAGLContext setCurrentContext:context];
    
	// This application only creates a single default framebuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple framebuffers.
    //glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
    glVertexPointer(3, GL_FLOAT, 0, vertices);
    glEnableClientState(GL_VERTEX_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, texCoords);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    
    glDrawArrays(GL_TRIANGLES, 0, NUM_VERTICES);
    
	// This application only creates a single color renderbuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple renderbuffers.
	glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{	
	// Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);

	glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
	glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight);

    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
	{
		NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
    
    return YES;
}

- (void) dealloc
{
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}

	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;

	if (shatterData) {
		free(shatterData);
		shatterData = nil;
	}
	if (vertices) {
		free(vertices);
		vertices = nil;
	}
	if (texCoords) {
		free(texCoords);
		texCoords = nil;
	}

	UIAccelerometer* accel = [UIAccelerometer sharedAccelerometer];
	accel.delegate = self;
	
	[super dealloc];
}

- (void) shatter:(CGPoint)point toggle:(BOOL)toggle
{
	if (toggle) {
		shattered = !shattered;
	}

	if (shattered) {
		CFAbsoluteTime time = CFAbsoluteTimeGetCurrent();
		if (shatteredTime == 0) {
			shatteredTime = time;
		}

#if TARGET_IPHONE_SIMULATOR
		double strength = 2.0 * random() / INT32_MAX * 10000;
#else
		double accelLen = sqrt(accelDlt[0]*accelDlt[0] + accelDlt[1]*accelDlt[1] + accelDlt[2]*accelDlt[2]);
		double strength = accelLen * 100000;
#endif

		for (int j = 0; j < 15; ++j) {
			for (int i = 0; i < 10; ++i) {
				int k = (j * 10 + i) * (3 + 3 + 3 + 1 + 1);
				
				if (time - shatteredTime < shatterData[k+10]) {
					GLfloat dx = (shatterData[k  ]*0.5+0.5)*backingWidth - point.x;
					GLfloat dy = (-shatterData[k+1]*0.5+0.5)*backingHeight - point.y;
					GLfloat dd = dx*dx + dy*dy;
					if (dd < strength) {
						shatterData[k+3] = ((random() / 2147483647.0) * 2 - 1) * 0.01;
						shatterData[k+4] = ((random() / 2147483647.0) * 2 - 1) * 0.01;
						//shatterData[k+5] = ((random() / 2147483647.0) * 2 - 1) * 0.01;
						shatterData[k+10] = dd / (strength*10);
					}
				}
			}
		}
	} else {
		shatteredTime = 0;
		[self resetShatterData];
	}
}

- (void)accelerometer:(UIAccelerometer *)accelerometer didAccelerate:(UIAcceleration *)acceleration
{
	accelDlt[0] = acceleration.x - accelVec[0];
	accelDlt[1] = acceleration.y - accelVec[1];
	accelDlt[2] = acceleration.z - accelVec[2];
	accelVec[0] = acceleration.x;
	accelVec[1] = acceleration.y;
	accelVec[2] = acceleration.z;
}

@end
